var __defProp = Object.defineProperty;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
var __async = (__this, __arguments, generator) => {
  return new Promise((resolve, reject) => {
    var fulfilled = (value) => {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    };
    var rejected = (value) => {
      try {
        step(generator.throw(value));
      } catch (e) {
        reject(e);
      }
    };
    var step = (x) => x.done ? resolve(x.value) : Promise.resolve(x.value).then(fulfilled, rejected);
    step((generator = generator.apply(__this, __arguments)).next());
  });
};

// src/event-emitter.ts
function createEventEmitter() {
  const listeners = /* @__PURE__ */ new Map();
  return {
    on(event, listener) {
      if (!listeners.has(event))
        listeners.set(event, /* @__PURE__ */ new Set());
      const set = listeners.get(event).add(listener);
      return {
        revoke: () => {
          set.delete(listener);
        }
      };
    },
    trigger(event, ...args) {
      if (!listeners.has(event))
        return;
      listeners.get(event).forEach((listener) => listener(...args));
    }
  };
}

// src/modifier-map.ts
function createModifierMap() {
  const modifiers2 = /* @__PURE__ */ new WeakMap();
  return {
    set(target, key, modifier, value) {
      let map;
      if (modifiers2.has(target))
        map = modifiers2.get(target);
      else
        modifiers2.set(target, map = /* @__PURE__ */ new Map());
      if (map.has(key))
        map.get(key)[modifier] = value;
      else
        map.set(key, { [modifier]: value });
    },
    get(target, key, modifier) {
      var _a, _b;
      return (_b = (_a = modifiers2.get(Object.getPrototypeOf(target))) == null ? void 0 : _a.get(key)) == null ? void 0 : _b[modifier];
    }
  };
}

// src/scheduler.ts
function createScheduler() {
  const queue = /* @__PURE__ */ new Set();
  const microtask = Promise.resolve();
  let flushing = false;
  const next = (cb) => {
    void microtask.then(cb);
  };
  const flush = () => {
    for (const task of queue) {
      task();
      queue.delete(task);
    }
  };
  return {
    flush,
    enqueue: (task) => {
      queue.add(task);
      if (!flushing) {
        next(() => {
          flushing = false;
          flush();
        });
      }
    }
  };
}

// src/model.ts
function createLink(target, key) {
  return [target, key];
}
function isEqualLink(x, y) {
  return x[0] === y[0] && x[1] === y[1];
}
var modifiers = createModifierMap();
function ignore(target, key) {
  modifiers.set(target, key, "i", true);
}
function effect(target, key) {
  modifiers.set(target, key, "ef", true);
}
function throttle(delay) {
  return (target, key) => {
    modifiers.set(target, key, "t", delay);
  };
}
function debounce(delay) {
  return (target, key) => {
    modifiers.set(target, key, "d", delay);
  };
}
function readonly(target, key) {
  modifiers.set(target, key, "r", true);
}
function extend(extender) {
  return (target, key) => {
    let extenders;
    if (extenders = modifiers.get(target, key, "e"))
      extenders.push(extender);
    else
      modifiers.set(target, key, "e", [extender]);
  };
}
var Model = class {
  constructor() {
    __publicField(this, "$schedule");
    __publicField(this, "$tick");
    __publicField(this, "$silent");
    __publicField(this, "$ref");
    __publicField(this, "$events", createEventEmitter());
    const scheduler = createScheduler();
    this.$schedule = scheduler.enqueue;
    this.$tick = scheduler.flush;
    const extend2 = (object, isRoot) => {
      let getRef = false;
      let silent = false;
      const getters = /* @__PURE__ */ new WeakMap();
      const computed = (target, key, receiver, descriptior) => {
        var _a;
        if (getters.has(descriptior.get))
          return getters.get(descriptior.get)();
        let value;
        let dirty = true;
        let dispose;
        let debounceHandle;
        const subs = /* @__PURE__ */ new Set();
        const extenders = (_a = modifiers.get(target, key, "e")) != null ? _a : [];
        const recompute = () => {
          dirty = false;
          let next;
          const observe2 = this.$magic(() => next = Reflect.get(target, key, receiver));
          if (next !== value) {
            value = next;
            for (const extender of extenders) {
              if (extender.compute)
                value = extender.compute(value);
            }
            dispose == null ? void 0 : dispose();
            dispose = observe2(() => __async(this, null, function* () {
              dirty = true;
              for (const extender of extenders) {
                if (extender.recompute && !(yield extender.recompute())) {
                  dirty = false;
                  return;
                }
                if (extender.notify && !(yield extender.notify(value)))
                  return;
              }
              const notify = () => {
                this.$events.trigger("set", createLink(target, key));
                for (const listener of subs)
                  listener();
              };
              let timeout;
              if ((timeout = modifiers.get(target, key, "t")) !== void 0) {
                setTimeout(notify, timeout);
              } else if ((timeout = modifiers.get(target, key, "d")) !== void 0) {
                if (debounceHandle !== void 0)
                  clearTimeout(debounceHandle);
                debounceHandle = setTimeout(notify, timeout);
              } else {
                this.$schedule(notify);
              }
            }));
          }
        };
        const get = () => {
          if (dirty)
            recompute();
          return value;
        };
        const observe = (listener) => {
          subs.add(listener);
          return {
            revoke() {
              subs.delete(listener);
            }
          };
        };
        for (const extender of extenders) {
          if (extender.init) {
            extender.init({
              get,
              observe,
              dispose() {
                throw new Error("cannot dispose");
              }
            });
          }
        }
        getters.set(descriptior.get, get);
        return get();
      };
      return new Proxy(object, {
        get: (target, key, receiver) => {
          const createReference = (link) => ({
            get: () => Reflect.get(...link),
            set: (value2) => void Reflect.set(...link, value2),
            observe: (listener) => this.$observe(link, listener)
          });
          if (isRoot) {
            if (key === "$ref") {
              getRef = true;
              return (link) => {
                return createReference(link);
              };
            }
            if (key === "$silent") {
              silent = true;
              return (value2) => {
                silent = false;
                return value2;
              };
            }
            if (typeof key === "string" && key.startsWith("$") || modifiers.get(target, key, "i") === true)
              return Reflect.get(target, key, receiver);
          }
          if (getRef) {
            getRef = false;
            return createLink(receiver, key);
          }
          const get = () => {
            const descriptior = Object.getOwnPropertyDescriptor(Object.getPrototypeOf(target), key);
            if (descriptior == null ? void 0 : descriptior.get)
              return computed(target, key, receiver, descriptior);
            return Reflect.get(target, key, receiver);
          };
          const value = get();
          if (!silent)
            this.$events.trigger("get", createLink(target, key));
          if (typeof value === "object" && value !== null)
            return extend2(value, false);
          return value;
        },
        set: (target, key, value, receiver) => {
          const succeeded = Reflect.set(target, key, value, receiver);
          this.$schedule(() => {
            this.$events.trigger("set", createLink(target, key));
          });
          return succeeded;
        }
      });
    };
    const extended = extend2(this, true);
    for (const key in this.constructor.prototype) {
      if (Object.prototype.hasOwnProperty.call(this.constructor.prototype, key)) {
        if (modifiers.get(this.constructor.prototype, key, "ef") === true)
          Reflect.get(extended, key);
      }
    }
    return extended;
  }
  $magic(callback) {
    const links = [];
    const temp = this.$events.on("get", (link) => links.push(link));
    callback();
    temp.revoke();
    return (listener) => {
      const subs = links.map((link) => this.$observe(link, listener));
      return () => subs.forEach((sub) => sub.revoke());
    };
  }
  $observe(target, listener) {
    return this.$events.on("set", (link) => {
      if (isEqualLink(target, link))
        listener();
    });
  }
  $computed(get) {
    const subscriptions = /* @__PURE__ */ new Set();
    let value;
    let dirty = true;
    let revoke;
    let disposed = false;
    const assertNotDisposed = () => {
      if (disposed)
        throw new Error("computed is disposed");
    };
    const recompute = () => {
      dirty = false;
      let next;
      const observe = this.$magic(() => next = get());
      if (next !== value) {
        value = next;
        revoke == null ? void 0 : revoke();
        revoke = observe(() => {
          dirty = true;
          this.$schedule(() => {
            for (const listener of subscriptions)
              listener();
          });
        });
      }
    };
    return {
      get: () => {
        assertNotDisposed();
        if (dirty)
          recompute();
        return value;
      },
      observe: (listener) => {
        assertNotDisposed();
        subscriptions.add(listener);
        return {
          revoke() {
            subscriptions.delete(listener);
          }
        };
      },
      dispose: () => {
        assertNotDisposed();
        disposed = true;
        revoke == null ? void 0 : revoke();
      }
    };
  }
};
export {
  Model,
  debounce,
  effect,
  extend,
  ignore,
  readonly,
  throttle
};
